@@ -1,10 +1,33 @@
+from __future__ import division
+
+__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"
+
+__license__ = """
+Permission is hereby granted, free of charge, to any person obtaining a copy
+of this software and associated documentation files (the "Software"), to deal
+in the Software without restriction, including without limitation the rights
+to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
+copies of the Software, and to permit persons to whom the Software is
+furnished to do so, subject to the following conditions:
+
+The above copyright notice and this permission notice shall be included in
+all copies or substantial portions of the Software.
+
+THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
+AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+THE SOFTWARE.
+"""
+
 import traits
 
 
-
-
 class Expression(object):
-    """An evaluatable part of a mathematical expression.
+    """Superclass for parts of a mathematical expression. Overrides operators
+    to implicitly construct :class:`Sum`, :class:`Product` and other expressions.
 
     Expression objects are immutable.
     """
@@ -127,23 +150,66 @@ class Expression(object):
         if not is_valid_operand(other):
             return NotImplemented
 
-        if is_zero(other): # exponent zero
+        if is_zero(other):  # exponent zero
             return 1
-        elif is_zero(other-1): # exponent one
+        elif is_zero(other-1):  # exponent one
             return self
         return Power(self, other)
 
     def __rpow__(self, other):
         assert is_constant(other)
 
-        if is_zero(other): # base zero
+        if is_zero(other):  # base zero
             return 0
-        elif is_zero(other-1): # base one
+        elif is_zero(other-1):  # base one
             return 1
         return Power(other, self)
 
     # }}}
 
+    # {{{ shifts
+
+    def __lshift__(self, other):
+        return LeftShift(self, other)
+
+    def __rlshift__(self, other):
+        return LeftShift(other, self)
+
+    def __rshift__(self, other):
+        return RightShift(self, other)
+
+    def __rrshift__(self, other):
+        return RightShift(other, self)
+
+    # }}}
+
+    # {{{ bitwise operators
+
+    def __inv__(self):
+        return BitwiseNot(self)
+
+    def __or__(self, other):
+        return BitwiseOr((self, other))
+
+    def __ror__(self, other):
+        return BitwiseOr((other, self))
+
+    def __xor__(self, other):
+        return BitwiseXor((self, other))
+
+    def __rxor__(self, other):
+        return BitwiseXor((other, self))
+
+    def __and__(self, other):
+        return BitwiseAnd((self, other))
+
+    def __rand__(self, other):
+        return BitwiseAnd((other, self))
+
+    # }}}
+
+    # {{{
+
     def __neg__(self):
         return -1*self
 
@@ -156,31 +222,48 @@ class Expression(object):
         else:
             return Subscript(self, subscript)
 
+    def attr(self, name):
+        """Return a :class:`Lookup` for *name* in *self*.
+        """
+        return Lookup(self, name)
+
     def __float__(self):
         from pymbolic.mapper.evaluator import evaluate_to_float
         return evaluate_to_float(self)
 
     def stringifier(self):
+        """Return a :class:`pymbolic.mapper.Mapper` class used to yield
+        a human-readable representation of *self*. Usually a subclass
+        of :class:`pymbolic.mapper.stringifier.StringifyMapper`.
+        """
         from pymbolic.mapper.stringifier import StringifyMapper
         return StringifyMapper
 
     def __str__(self):
+        """Use the :meth:`stringifier` to return a human-readable
+        string representation of *self*.
+        """
+
         from pymbolic.mapper.stringifier import PREC_NONE
         return self.stringifier()()(self, PREC_NONE)
 
     def __repr__(self):
+        """Provides a default :func:`repr` based on
+        the Python pickling interface :meth:`__getinitargs__`.
+        """
         initargs_str = ", ".join(repr(i) for i in self.__getinitargs__())
 
         return "%s(%s)" % (self.__class__.__name__, initargs_str)
+    # }}}
 
-    # {{{ hashable interface
+    # {{{ hash/equality interface
 
     def __eq__(self, other):
         """Provides equality testing with quick positive and negative paths
-        based on L{id} and L{__hash__}().
+        based on :func:`id` and :meth:`__hash__`.
 
-        Subclasses should generally not override this method, but instead 
-        provide an implementation of L{is_equal}.
+        Subclasses should generally not override this method, but instead
+        provide an implementation of :meth:`is_equal`.
         """
         if self is other:
             return True
@@ -195,8 +278,8 @@ class Expression(object):
     def __hash__(self):
         """Provides caching for hash values.
 
-        Subclasses should generally not override this method, but instead 
-        provide an implementation of L{get_hash}.
+        Subclasses should generally not override this method, but instead
+        provide an implementation of :meth:`get_hash`.
         """
         try:
             return self.hash_value
@@ -206,33 +289,36 @@ class Expression(object):
 
     # }}}
 
-    # {{{ hashable backend
+    # {{{ hash/equality backend
 
     def is_equal(self, other):
         return (type(other) == type(self)
                 and self.__getinitargs__() == other.__getinitargs__())
 
     def get_hash(self):
-        return hash((type(self),)+ self.__getinitargs__())
+        return hash((type(self).__name__,) + self.__getinitargs__())
 
     # }}}
 
-    # {{{ comparison interface
+    # {{{ prevent less / greater comparisons
 
     # /!\ Don't be tempted to resolve these to ComparisonOperator.
 
-    def __le__(self, other): raise TypeError("expressions don't have an order")
-    def __lt__(self, other): raise TypeError("expressions don't have an order")
-    def __ge__(self, other): raise TypeError("expressions don't have an order")
-    def __gt__(self, other): raise TypeError("expressions don't have an order")
+    def __le__(self, other):
+        raise TypeError("expressions don't have an order")
+
+    def __lt__(self, other):
+        raise TypeError("expressions don't have an order")
+
+    def __ge__(self, other):
+        raise TypeError("expressions don't have an order")
+
+    def __gt__(self, other):
+        raise TypeError("expressions don't have an order")
 
     # }}}
 
 
-
-
-
-
 class AlgebraicLeaf(Expression):
     """An expression that serves as a leaf for arithmetic evaluation.
     This may end up having child nodes still, but they're not reached by
@@ -240,17 +326,16 @@ class AlgebraicLeaf(Expression):
     pass
 
 
-
-
 class Leaf(AlgebraicLeaf):
-    """An expression that is irreducible, i.e. has no Expression-type parts 
+    """An expression that is irreducible, i.e. has no Expression-type parts
     whatsoever."""
     pass
 
 
-
-
 class Variable(Leaf):
+    """
+    .. attribute:: name
+    """
     def __init__(self, name):
         self.name = name
 
@@ -266,8 +351,6 @@ class Variable(Leaf):
     mapper_method = intern("map_variable")
 
 
-
-
 class Wildcard(Leaf):
     def __getinitargs__(self):
         return ()
@@ -288,13 +371,21 @@ class FunctionSymbol(AlgebraicLeaf):
     mapper_method = intern("map_function_symbol")
 
 
-
-
-
 # {{{ structural primitives
 
 class Call(AlgebraicLeaf):
+    """A function invocation.
+
+    .. attribute:: function
+    .. attribute:: parameters
+    """
+
     def __init__(self, function, parameters):
+        """
+        :arg function: A :class:`Expression` that evaluates to a function.
+        :arg parameters: A :class:`tuple` of positional paramters.
+        """
+
         self.function = function
         self.parameters = parameters
 
@@ -314,9 +405,12 @@ class Call(AlgebraicLeaf):
     mapper_method = intern("map_call")
 
 
-
-
 class Subscript(AlgebraicLeaf):
+    """An array subscript.
+
+    .. attribute:: aggregate
+    .. attribute:: index
+    """
     def __init__(self, aggregate, index):
         self.aggregate = aggregate
         self.index = index
@@ -327,9 +421,11 @@ class Subscript(AlgebraicLeaf):
     mapper_method = intern("map_subscript")
 
 
-
-
 class Lookup(AlgebraicLeaf):
+    """Access to an attribute of an *aggregate*, such as an
+    attribute of a class.
+    """
+
     def __init__(self, aggregate, name):
         self.aggregate = aggregate
         self.name = name
@@ -341,9 +437,10 @@ class Lookup(AlgebraicLeaf):
 
 # }}}
 
+
 # {{{ arithmetic primitives
 
-class Sum(Expression):
+class _MultiChildExpression(Expression):
     def __init__(self, children):
         assert isinstance(children, tuple)
 
@@ -352,6 +449,14 @@ class Sum(Expression):
     def __getinitargs__(self):
         return self.children
 
+
+class Sum(_MultiChildExpression):
+    """
+    .. attribute:: children
+
+        A :class:`tuple`.
+    """
+
     def __add__(self, other):
         if not is_valid_operand(other):
             return NotImplemented
@@ -392,15 +497,12 @@ class Sum(Expression):
     mapper_method = intern("map_sum")
 
 
+class Product(_MultiChildExpression):
+    """
+    .. attribute:: children
 
-
-class Product(Expression):
-    def __init__(self, children):
-        assert isinstance(children, tuple)
-        self.children = children
-
-    def __getinitargs__(self):
-        return self.children
+        A :class:`tuple`.
+    """
 
     def __mul__(self, other):
         if not is_valid_operand(other):
@@ -433,8 +535,6 @@ class Product(Expression):
     mapper_method = intern("map_product")
 
 
-
-
 class QuotientBase(Expression):
     def __init__(self, numerator, denominator=1):
         self.numerator = numerator
@@ -455,9 +555,12 @@ class QuotientBase(Expression):
         return bool(self.numerator)
 
 
-
-
 class Quotient(QuotientBase):
+    """
+    .. attribute:: numerator
+    .. attribute:: denominator
+    """
+
     def is_equal(self, other):
         from pymbolic.rational import Rational
         return isinstance(other, (Rational, Quotient)) \
@@ -467,21 +570,30 @@ class Quotient(QuotientBase):
     mapper_method = intern("map_quotient")
 
 
-
-
 class FloorDiv(QuotientBase):
+    """
+    .. attribute:: numerator
+    .. attribute:: denominator
+    """
+
     mapper_method = intern("map_floor_div")
 
 
-
-
 class Remainder(QuotientBase):
+    """
+    .. attribute:: numerator
+    .. attribute:: denominator
+    """
+
     mapper_method = intern("map_remainder")
 
 
-
-
 class Power(Expression):
+    """
+    .. attribute:: base
+    .. attribute:: exponent
+    """
+
     def __init__(self, base, exponent):
         self.base = base
         self.exponent = exponent
@@ -493,14 +605,105 @@ class Power(Expression):
 
 # }}}
 
+
+# {{{ shift operators
+
+class _ShiftOperator(Expression):
+    def __init__(self, shiftee, shift):
+        self.shiftee = shiftee
+        self.shift = shift
+
+    def __getinitargs__(self):
+        return self.base, self.exponent
+
+
+class LeftShift(Expression):
+    """
+    .. attribute:: shiftee
+    .. attribute:: shift
+    """
+
+    mapper_method = intern("map_left_shift")
+
+
+class RightShift(Expression):
+    """
+    .. attribute:: shiftee
+    .. attribute:: shift
+    """
+
+    mapper_method = intern("map_right_shift")
+
+# }}}
+
+
+# {{{ bitwise operators
+
+class BitwiseNot(Expression):
+    """
+    .. attribute:: child
+    """
+
+    def __init__(self, child):
+        self.child = child
+
+    def __getinitargs__(self):
+        return (self.child,)
+
+    mapper_method = intern("map_bitwise_not")
+
+
+class BitwiseOr(_MultiChildExpression):
+    """
+    .. attribute:: children
+
+        A :class:`tuple`.
+    """
+
+    mapper_method = intern("map_bitwise_or")
+
+
+class BitwiseXor(_MultiChildExpression):
+    """
+    .. attribute:: children
+
+        A :class:`tuple`.
+    """
+
+    mapper_method = intern("map_bitwise_xor")
+
+
+class BitwiseAnd(_MultiChildExpression):
+    """
+    .. attribute:: children
+
+        A :class:`tuple`.
+    """
+
+    mapper_method = intern("map_bitwise_and")
+
+# }}}
+
+
 # {{{ comparisons, logic, conditionals
 
 class ComparisonOperator(Expression):
-    """Note: comparisons are not implicitly constructed by comparing
-    Expression objects.
+    """
+    .. attribute:: left
+    .. attribute:: operator
+    .. attribute:: right
+
+    .. note::
+
+        Unlike other expressions, comparisons are not implicitly constructed by
+        comparing :class:`Expression` objects.
     """
 
     def __init__(self, left, operator, right):
+        """
+        :arg operator: One of ``[">", ">=", "==", "!=", "<", "<="]``.
+        """
+
         self.left = left
         self.right = right
         if not operator in [">", ">=", "==", "!=", "<", "<="]:
@@ -513,52 +716,46 @@ class ComparisonOperator(Expression):
     mapper_method = intern("map_comparison")
 
 
+class LogicalNot(Expression):
+    """
+    .. attribute:: child
+    """
 
-
-class BooleanExpression(Expression):
-    pass
-
-class LogcialNot(BooleanExpression):
     def __init__(self, child):
         self.child = child
 
     def __getinitargs__(self):
-        return (self.child, self.prefix)
+        return (self.child,)
 
     mapper_method = intern("map_logical_not")
 
 
+class LogicalOr(_MultiChildExpression):
+    """
+    .. attribute:: children
 
-
-class LogicalOr(BooleanExpression):
-    def __init__(self, children):
-        assert isinstance(children, tuple)
-
-        self.children = children
-
-    def __getinitargs__(self):
-        return self.children
+        A :class:`tuple`.
+    """
 
     mapper_method = intern("map_logical_or")
 
 
+class LogicalAnd(_MultiChildExpression):
+    """
+    .. attribute:: children
 
-
-class LogicalAnd(BooleanExpression):
-    def __init__(self, children):
-        assert isinstance(children, tuple)
-
-        self.children = children
-
-    def __getinitargs__(self):
-        return self.children
+        A :class:`tuple`.
+    """
 
     mapper_method = intern("map_logical_and")
 
 
-
-
 class If(Expression):
+    """
+    .. attribute:: condition
+    .. attribute:: then
+    .. attribute:: else_
+    """
     def __init__(self, criterion, then, else_):
         self.condition = criterion
         self.then = then
@@ -570,8 +767,6 @@ class If(Expression):
     mapper_method = intern("map_if")
 
 
-
-
 class IfPositive(Expression):
     def __init__(self, criterion, then, else_):
         from warnings import warn
@@ -588,8 +783,6 @@ class IfPositive(Expression):
     mapper_method = intern("map_if_positive")
 
 
-
-
 class _MinMaxBase(Expression):
     def __init__(self, children):
         self.children = children
@@ -597,15 +790,18 @@ class _MinMaxBase(Expression):
     def __getinitargs__(self):
         return self.children
 
+
 class Min(_MinMaxBase):
     mapper_method = intern("map_min")
 
+
 class Max(_MinMaxBase):
     mapper_method = intern("map_max")
 
 # }}}
 
-# {{{
+
+# {{{ misc stuff
 
 class Vector(Expression):
     """An immutable sequence that you can compute with."""
@@ -634,22 +830,22 @@ class Vector(Expression):
 
     def __add__(self, other):
         if len(other) != len(self):
-            raise ValueError, "can't add values of differing lengths"
+            raise ValueError("can't add values of differing lengths")
         return Vector(tuple(x+y for x, y in zip(self, other)))
 
     def __radd__(self, other):
         if len(other) != len(self):
-            raise ValueError, "can't add values of differing lengths"
+            raise ValueError("can't add values of differing lengths")
         return Vector(tuple(y+x for x, y in zip(self, other)))
 
     def __sub__(self, other):
         if len(other) != len(self):
-            raise ValueError, "can't subtract values of differing lengths"
+            raise ValueError("can't subtract values of differing lengths")
         return Vector(tuple(x-y for x, y in zip(self, other)))
 
     def __rsub__(self, other):
         if len(other) != len(self):
-            raise ValueError, "can't subtract values of differing lengths"
+            raise ValueError("can't subtract values of differing lengths")
         return Vector(tuple(y-x for x, y in zip(self, other)))
 
     def __mul__(self, other):
@@ -675,15 +871,57 @@ class Vector(Expression):
     mapper_method = intern("map_vector")
 
 
+class cse_scope:
+    """Determines the lifetime for the saved value of a :class:`CommonSubexpression`.
+
+    .. attribute:: EVALUATION
+
+        The evaluated result lives for the duration of the evaluation of the
+        current expression and is discarded thereafter.
+
+    .. attribute:: EXPRESSION
+
+        The evaluated result lives for the lifetime of the current expression
+        (across multiple evaluations with multiple parameters) and is discarded
+        when the expression is.
+
+    .. attribute:: GLOBAL
+
+        The evaluated result lives until the execution context dies.
+    """
+
+    EVALUATION = "pymbolic_eval"
+    EXPRESSION = "pymbolic_expr"
+    GLOBAL = "pymbolic_global"
 
 
 class CommonSubexpression(Expression):
-    def __init__(self, child, prefix=None):
+    """A helper for code generation and caching. Denotes a subexpression that
+    should only be evaluated once. If, in code generation, it is assigned to
+    a variable, a name starting with :attr:`prefix` should be used.
+
+    .. attribute:: child
+    .. attribute:: prefix
+    .. attribute:: scope
+
+        One of the values in :class:`cse_scope`. See there for meaning.
+
+    See :class:`pymbolic.mapper.c_code.CCodeMapper` for an example.
+    """
+
+    def __init__(self, child, prefix=None, scope=None):
+        """
+        :arg scope: Defaults to :attr:`cse_scope.EVALUATION` if given as *None*.
+        """
+        if scope is None:
+            scope = cse_scope.EVALUATION
+
         self.child = child
         self.prefix = prefix
+        self.scope = scope
 
     def __getinitargs__(self):
-        return (self.child, self.prefix)
+        return (self.child, self.prefix, self.scope)
 
     def get_extra_properties(self):
         """Return a dictionary of extra kwargs to be passed to the
@@ -698,7 +936,6 @@ class CommonSubexpression(Expression):
     mapper_method = intern("map_common_subexpression")
 
 
-
 class Substitution(Expression):
     """Work-alike of sympy's Subs."""
 
@@ -713,8 +950,6 @@ class Substitution(Expression):
     mapper_method = intern("map_substitution")
 
 
-
-
 class Derivative(Expression):
     """Work-alike of sympy's Derivative."""
 
@@ -728,8 +963,6 @@ class Derivative(Expression):
     mapper_method = intern("map_derivative")
 
 
-
-
 class Slice(Expression):
     """A slice expression as in a[1:7]."""
 
@@ -773,7 +1006,9 @@ class Slice(Expression):
 
 # }}}
 
-# intelligent factory functions ----------------------------------------------
+
+# {{{ intelligent factory functions
+
 def make_variable(var_or_string):
     if not isinstance(var_or_string, Expression):
         return Variable(var_or_string)
@@ -781,14 +1016,10 @@ def make_variable(var_or_string):
         return var_or_string
 
 
-
-
 def subscript(expression, index):
     return Subscript(expression, index)
 
 
-
-
 def flattened_sum(components):
     # flatten any potential sub-sums
     queue = list(components)
@@ -813,16 +1044,12 @@ def flattened_sum(components):
         return Sum(tuple(done))
 
 
-
-
 def linear_combination(coefficients, expressions):
     return sum(coefficient * expression
                  for coefficient, expression in zip(coefficients, expressions)
                  if coefficient and expression)
 
 
-
-
 def flattened_product(components):
     # flatten any potential sub-products
     queue = list(components)
@@ -849,14 +1076,6 @@ def flattened_product(components):
         return Product(tuple(done))
 
 
-
-
-def polynomial_from_expression(expression):
-    pass
-
-
-
-
 def quotient(numerator, denominator):
     if not (denominator-1):
         return numerator
@@ -877,10 +1096,11 @@ def quotient(numerator, denominator):
 
     return Quotient(numerator, denominator)
 
+# }}}
 
 
+# {{{ tool functions
 
-# {{{ tool functions --------------------------------------------------------------
 global VALID_CONSTANT_CLASSES
 global VALID_OPERANDS
 VALID_CONSTANT_CLASSES = (int, float, complex)
@@ -893,22 +1113,20 @@ except ImportError:
     pass
 
 
-
-
 def is_constant(value):
     return isinstance(value, VALID_CONSTANT_CLASSES)
 
+
 def is_valid_operand(value):
     return isinstance(value, VALID_OPERANDS) or is_constant(value)
 
 
-
-
 def register_constant_class(class_):
     global VALID_CONSTANT_CLASSES
 
     VALID_CONSTANT_CLASSES += (class_,)
 
+
 def unregister_constant_class(class_):
     global VALID_CONSTANT_CLASSES
 
@@ -917,20 +1135,17 @@ def unregister_constant_class(class_):
     VALID_CONSTANT_CLASSES = tuple(tmp)
 
 
-
-
 def is_nonzero(value):
     try:
         return bool(value)
     except ValueError:
         return True
 
+
 def is_zero(value):
     return not is_nonzero(value)
 
 
-
-
 def wrap_in_cse(expr, prefix=None):
     if isinstance(expr, (Variable, Subscript)):
         return expr
@@ -948,10 +1163,23 @@ def wrap_in_cse(expr, prefix=None):
         return CommonSubexpression(expr, prefix)
 
 
+def make_common_subexpression(field, prefix=None, scope=None):
+    """Wrap *field* in a :class:`CommonSubexpression` with
+    *prefix*. If *field* is a :mod:`numpy` object array,
+    each individual entry is instead wrapped. If *field* is a
+    :class:`pymbolic.geometric_algebra.MultiVector`, each
+    coefficient is individually wrapped.
 
+    See :class:`CommonSubexpression` for the meaning of *prefix*
+    and *scope*.
+    """
 
+    if isinstance(field, CommonSubexpression) and (
+            scope is None or scope == cse_scope.EVALUATION
+            or field.scope == scope):
+        # Don't re-wrap
+        return field
 
-def make_common_subexpression(field, prefix=None):
     try:
         from pytools.obj_array import log_shape
     except ImportError:
@@ -962,7 +1190,22 @@ def make_common_subexpression(field, prefix=None):
     if have_obj_array:
         ls = log_shape(field)
 
-    if have_obj_array and ls != ():
+    from pymbolic.geometric_algebra import MultiVector
+    if isinstance(field, MultiVector):
+        new_data = {}
+        for bits, coeff in field.data.iteritems():
+            if prefix is not None:
+                blade_str = field.space.blade_bits_to_str(bits, "")
+                component_prefix = prefix+"_"+blade_str
+            else:
+                component_prefix = None
+
+            new_data[bits] = make_common_subexpression(
+                    coeff, component_prefix, scope)
+
+        return MultiVector(new_data, field.space)
+
+    elif have_obj_array and ls != ():
         from pytools import indices_in_shape
         result = numpy.zeros(ls, dtype=object)
 
@@ -975,21 +1218,20 @@ def make_common_subexpression(field, prefix=None):
             if is_constant(field[i]):
                 result[i] = field[i]
             else:
-                result[i] = CommonSubexpression(field[i], component_prefix)
+                result[i] = make_common_subexpression(
+                        field[i], component_prefix, scope)
 
         return result
     else:
         if is_constant(field):
             return field
         else:
-            return CommonSubexpression(field, prefix)
-
-
+            return CommonSubexpression(field, prefix, scope)
 
 
 def make_sym_vector(name, components):
     """Return an object array of *components* subscripted
-    :class:`Field` instances.
+    :class:`Variable` instances.
 
     :param components: The number of components in the vector.
     """
@@ -1001,14 +1243,27 @@ def make_sym_vector(name, components):
     return join_fields(*[vfld[i] for i in components])
 
 
+def make_sym_array(name, shape):
+    vfld = Variable(name)
+    if shape == ():
+        return vfld
+
+    import numpy as np
+    result = np.zeros(shape, dtype=object)
+    from pytools import indices_in_shape
+    for i in indices_in_shape(shape):
+        result[i] = vfld[i]
+
+    return result
 
 
 def variables(s):
+    """Return a list of variables for each (space-delimited) identifier
+    in *s*.
+    """
     return [Variable(s_i) for s_i in s.split() if s_i]
 
 # }}}
 
 
-
-
 # vim: foldmethod=marker
