@@ -27,6 +27,23 @@ import pytest
 
 from pymbolic.mapper import IdentityMapper
 
+try:
+    reduce
+except NameError:
+    from functools import reduce
+
+
+def test_integer_power():
+    from pymbolic.algorithm import integer_power
+
+    for base, expn in [
+            (17, 5),
+            (17, 2**10),
+            (13, 20),
+            (13, 1343),
+            ]:
+        assert base**expn == integer_power(base, expn)
+
 
 def test_expand():
     from pymbolic import var, expand
@@ -87,12 +104,14 @@ def test_sympy_interaction():
     p2s = PymbolicToSympyMapper()
 
     p1_expr = s2p(s1_expr)
-
     s2_expr = p2s(p1_expr)
-    assert s1_expr == s2_expr
+
+    assert sp.ratsimp(s1_expr - s2_expr) == 0
 
     p2_expr = s2p(s2_expr)
-    assert p1_expr == p2_expr
+    s3_expr = p2s(p2_expr)
+
+    assert sp.ratsimp(s1_expr - s3_expr) == 0
 
 
 # {{{ fft
